def invertebrate(pos):
    return """InvertEn(pos=%s, rel=rel, angle = angle)""" % (repr(pos),)

def moving_replace(pos, dir, time, o1, o2):
    # FIXME: we change params for o1 and o2 that require knowledge of
    # Shape internals. This is a huge hack!
    newpos = pos[0]+time*dir[0], pos[1]+time*dir[1]
    o2.pos = newpos
    if not o2.pattern:
        o2.pattern = [(dir, 0)]
    pattern=Pattern([(dir, time), 'delete(self)', 'place(new)'], new=o2)
    o1.pos = pos
    o1.pattern = pattern
    o1.kwargs['static'] = True
    place(o1)
    return o2

def place_inverts(rel, n, axis, posx, dirx, starty, diffy, startn=0):
    '''Place n invertebrates with typical invertebrate animations.
    
    If axis is 1, place invertebrates at (posx, starty+i*diffy). They
    will move towards (posx+dirx*30, starty+i*diffy).

    If axis is 0, place them at (starty+i*diffy, posx). They will move
    towards (starty+i*diffy, posx+dirx*30).

    Returns a list of enemies.'''    
    room1es = []
    for i in range(n):
        py = starty+diffy*i
        if axis == 1:
            move = ((3*dirx, 0), 10)
            cpos = [posx, py]
            epos = [30*dirx+posx, py]
            if dirx > 0:
                angle = 0
            else:
                angle = 180
        else:
            move = ((0, 3*dirx), 10)
            cpos = [py, posx]
            epos = [py, 30*dirx+posx]
            if dirx > 0:
                angle = 270
            else:
                angle = 90
        pattern = Pattern([((0, 0), 10 + 5*(i+startn)), move, 'delete(self)', 'place(room1es[%d])'%i], room1es=room1es)
        c = circle(14, static=True, pos=cpos, pattern=pattern, image='invertebrate-bg', angle=angle, rel=rel)
        exec "new = %s" % invertebrate(epos)
        room1es.append(new)
    # 10+5*i waiting, then moving for 10 frames, then one extra frame
    # to place the enemy. i ranges 0..n-1.
    t = 10+5*(startn+n-1)+10+1
    return room1es

class ShipAssault(object):
    def __init__(self, outside, packpos, packside = 'top', payloadtype=None, doorlabel='indoor'):
        packpos = list(packpos)
        axis = {'top': 1, 'bottom': 1, 'left': 0, 'right': 0}[packside]
        direction = {'top': -1, 'bottom': 1, 'left': -1, 'right': 1}[packside]
        angles = {'top': -90, 'bottom': 90, 'left': 0, 'right': 180}
        packpos[axis] += 1000*direction
        movedir = [0,0]
        movedir[axis] -= 10*direction
        self.door = door(outside, doorlabel)
        dside = opposites[packside]
        angle = angles[packside]
        if payloadtype == None:
            payloadtype = Player
        ppack, p = hubox(payload=payloadtype(angle=angle), rel=(packside, outside),
                           pos=packpos,
                           pattern=Pattern([(tuple(movedir), 100), 'd.open()',
                                    'delete(self, %s)'%repr(dside), ((0, 0), 0),
                                    'set_flag("arrived", True)'], d=self.door))
        self.package = ppack
        self.player = p

def enemyship(*args, **kwargs):
    ship = ShipAssault(*args, **kwargs)
    return ship.player

def make_enemy_every(rel, x, y, ticks=30, axis=1, dirx=1):
    cpos = (x, y)
    if axis == 1:
        move = ((3*dirx, 0), 10)
        epos = [30*dirx+x, y]
        if dirx > 0:
            angle = 0
        else:
            angle = 180
    else:
        move = ((0, 3*dirx), 10)
        epos = [x, 30*dirx+y]
        if dirx > 0:
            angle = 270
        else:
            angle = 90

    new = invertebrate(epos)
    placecmd = "place(%s)" % new
    pattern = "Pattern([%s, 'delete(self)', %s], rel=rel, angle=angle)"%(repr(move), repr(placecmd))
    c = "circle(14, static=True, pos=%s, pattern=%s, image='invertebrate-bg', angle=angle)"% (repr(cpos), pattern)
    return Pattern([Repeat, c, Ticks(ticks)], rel=rel, angle=angle)

class SwitchRepeat(Scenery):
    '''Add a repeating action which blocks on flag, then moves a switch'''
    def __init__(self, initfunc, func, flag, dir=None, goals=[], type=Switch, pos=None, **kwargs):
        DIST = 20
        pos = list(pos)
        pos[0] += dir[0]*DIST
        pos[1] += dir[1]*DIST
        sflag = 'switch'
        kwargs['action'] = ["set_flag(%s, True)"%repr(sflag)]
        self.switch = type(pos=pos, **kwargs)
        # Sometimes this switch gets hit twice in a row, leading to
        # sflag getting set to True after the pattern that checks for sflag
        # runs, which leads to sflag being True by the time we get to initfunc.
        # This leads to func being called twice and self being
        # in a weird location.
        # For this reason, we set sflag=False after we block on flag,
        # and only just before we move into location.
        # Contrariwise, we don't set the given flag's value until
        # after our own flag is True, so that the programmer can check against
        # the flag's value while the button hasn't been pushed.
        self.action2 = [Flag(flag, True),
                        'set_flag(%s, False)'%repr(sflag),
                        initfunc,
                        'start(s, [(%s, 10), ((0, 0), 0)])'%((dir[0]*-2, dir[1]*-2),),
                        Flag(sflag, True),
                        'set_flag(%s, None)'%repr(flag),
                        All(goals),
                        'start(s, [(%s, 10), ((0, 0), 0)])'%((dir[0]*2, dir[1]*2),),
                        func,
                        Repeat]

    def place(self, game):
        add_action(s=self.switch, *self.action2)
        return place(self.switch)

def switchrepeat(*args, **kwargs):
    return place(SwitchRepeat(*args, **kwargs))

class HeartObj(Shape):
    def __init__(self, roomargs, flipv):
        rightpat = [((5, 0), 40), ((0, 0), 0), ((-5, 0), 40), ((0, 0), 0), Repeat]
        # rightpat[2:2] = ['if get_flag("bigedie") == 1: print "stop"; stop(bige.diesound)', 'if get_flag("bigedie"): set_flag("bigedie", 1)']
        leftpat = mirror_x(rightpat)
        if flipv:
            axis = -1
        else:
            axis = 1
        roompts = [(600, 0), (600, axis*200), (0, axis*200), (0, 0)]
        heartroom = enwalls(roompts, counter=flipv,
                             use_last_end=True, use_first_start=True,
                             last_end=flip('enemywall-tee', False),
                             first_start='enemywall-tee',
                             **roomargs)
        if not flipv:
            leftargs = {'use_startcorner': True}
            rightargs = {'use_endcorner': True,
                         'endcorner': flip('enemywall-tee', fliph=True)}
        else:
            rightargs = {'use_startcorner': True}
            leftargs = {'use_endcorner': True,
                        'endcorner': flip('enemywall-tee', fliph=True)}

        leftw = enwalls([(0, axis*200), 'w', (0, 0), (-200, 0)], pattern=rightpat, rel=heartroom, pos=(100, 0), static=True, counter=flipv, under=True,
                        **leftargs)
        chain(heartroom, leftw, (0, axis*100), (-100, axis*100), texture="heartpiston")
        image(leftw, 'w', 'left', image='heartwall', layer=LAYERGEOMS, flipv=flipv)

        rightw = enwalls([(200, 0), (0, 0), 'w', (0, axis*200)], pattern=leftpat, rel=heartroom, pos=(500, 0), static=True, counter=flipv, under=True,
                         **rightargs)
        chain(rightw, heartroom, (0, axis*100), (600, axis*100), texture="heartpiston")
        image(rightw, 'w', 'right', image='heartwall', fliph=True, layer=LAYERGEOMS, flipv=flipv)

        self.leftw = leftw
        self.rightw = rightw

    def stop(self):
        stoppat = [((0, 0), 0)]
        start(self.leftw, stoppat)
        start(self.rightw, stoppat)

def heart(roomargs, flipv=False):
    '''Hearts are made of two sections of 200 units wide each. These retract
    to make a space 400 units wide -- so altogether the room needs to be at
    least 800 units wide. Would smaller hearts be good? Different speeds? Not sure.'''
    return HeartObj(roomargs, flipv)

def allytype(*args, **kwargs):
    kwargs.setdefault('health',100)
    kwargs.setdefault('side', SIDE_PLAYER)
    #kwargs.setdefault('style', 'human')
    return Controlled(shapes=[Circle(16, image='ship')], *args, **kwargs)

class KillAll(All):
    def __init__(self, lst):
        super(KillAll, self).__init__(map(Kills, lst))

def killall(*args, **kwargs):
    return wait(KillAll(*args, **kwargs))
