class Event(object):
    """this is a superclass for any events that might be generated by an
    object and sent to the EventManager"""
    def __init__(self):
        self.name = "Generic Event"

class ObjectRefreshEvent(Event):
    def __init__(self, obj):
        self.obj = obj

class ObjectDeleteEvent(Event):
    def __init__(self, obj, killstr=None, remains=False, chrome=True):
        # set chrome to False to disable corpses and whatnot
        self.obj = obj
        self.killstr = killstr
        self.remains = remains
        self.chrome = chrome

class ObjectKillEvent(ObjectDeleteEvent):
    pass

class ObjectDestroyPartEvent(Event):
    def __init__(self, obj):
        self.obj = obj

class ObjectHitEvent(Event):
    def __init__(self, obj, killstr, dies=False, geom=None, strength=None):
        self.obj = obj
        self.killstr = killstr
        self.dies = dies
        self.geom = geom
        self.strength = strength

class ObjectIncomingEvent(Event):
    def __init__(self, obj):
        self.obj = obj

class ExplosionSpriteEvent(Event):
    def __init__(self, *args):
        self.args = args

class AddImageEvent(Event):
    def __init__(self, layer, pos, strip, rot=0, attr=None, rel=None, start=None, texture=None, repeat=True, attach=None, merge=None, type=None, **kwargs):
        self.layer = layer
        self.pos = pos
        self.strip = strip
        self.rot = rot
        self.attr = attr
        self.rel = rel
        self.texture = texture
        self.start = start
        self.repeat = repeat
        self.attach = attach
        self.merge = merge
        self.type = type
        self.kwargs = kwargs

class NewGeomEvent(Event):
    def __init__(self, geom):
        self.geom = geom

class NewObjectEvent(Event):
    def __init__(self, obj):
        self.obj = obj

class ToggleGeomEvent(Event):
    def __init__(self, geom, toggle):
        self.geom = geom
        self.toggle = toggle

class AddChainEvent(Event):
    def __init__(self, obj1, obj2, pos1, pos2, txt, rot, start=0):
        self.obj1 = obj1
        self.obj2 = obj2
        self.pos1 = pos1
        self.pos2 = pos2
        self.txt = txt
        self.rot = rot
        self.start = start

class GeomImageEvent(Event):
    def __init__(self, geom, image=None, texture=None, attach=True,
                 layer=None):
        self.geom = geom
        self.image = image
        self.texture = texture
        self.attach = attach
        self.layer = layer

class FXEvent(Event):
    def __init__(self, part, type, kwargs):
        self.part = part
        self.type = type
        self.kwargs = kwargs

class ObjectShootingEvent(Event):
    def __init__(self, obj):
        self.obj = obj

class NewSoundEvent(Event):
    def __init__(self, filename, loop=False, music=False, vol=1.0, source=None):
        self.file = filename
        self.source = source
        self.loop = loop
        self.music = music
        self.vol = vol

class StopSoundEvent(Event):
    def __init__(self, filename):
        self.file = filename

class ScreenshotEvent(Event):
    pass

class RenderAllEvent(Event):
    pass

class PauseEvent(Event):
    pass

class ConfigChangedEvent(Event):
    def __init__(self, key):
        self.key = key
